@@ -11,7 +11,7 @@ region.
 '''
 
 from ast import literal_eval
-from collections import OrderedDict
+from collections import OrderedDict, defaultdict
 from datetime import datetime
 from decimal import Decimal
 from functools import partial, partialmethod
@@ -112,6 +112,13 @@ class EpochAxisItem(pg.AxisItem):
         years_indices = [ceil(yi) for yi in _pdtime2index(ax, years)]
         return [(0,years_indices)]
 
+    def generateDrawSpecs(self, p):
+        specs = super().generateDrawSpecs(p)
+        if specs and not self.style['showValues']:
+            pen,p0,p1 = specs[0]
+            specs = [(_makepen('#fff0'),p0,p1)] + list(specs[1:]) # don't draw axis if hiding values
+        return specs
+
 
 
 class YAxisItem(pg.AxisItem):
@@ -188,6 +195,7 @@ class PandasDataSource:
         self.df.columns = newcols
         self.pre_update = lambda df: df
         self._period = None
+        self.is_sparse = self.df[self.df.columns[self.col_data_offset]].isnull().sum() > len(self.df)/2
 
     @property
     def period(self):
@@ -283,6 +291,9 @@ class PandasDataSource:
         self.cache_hilo = OrderedDict()
         self._period = None
         datasrc._period = None
+        ldf2 = len(self.df) / 2
+        self.is_sparse = self.is_sparse or self.df[self.df.columns[self.col_data_offset]].isnull().sum() > ldf2
+        datasrc.is_sparse = datasrc.is_sparse or datasrc.df[datasrc.df.columns[datasrc.col_data_offset]].isnull().sum() > ldf2
 
     def update(self, datasrc):
         df = self.pre_update(self.df)
@@ -291,6 +302,8 @@ class PandasDataSource:
         df = df.set_index(timecol)
         input_df = datasrc.df.set_index(datasrc.df.columns[0])
         input_df.columns = [self.renames.get(col, col) for col in input_df.columns]
+        # pad index if the input data is a sub-set
+        input_df = pd.merge(input_df, df[[]], how='outer', left_index=True, right_index=True)
         for col in df.columns:
             if col not in input_df.columns:
                 input_df[col] = df[col]
@@ -335,6 +348,11 @@ class PandasDataSource:
 
     def rows(self, colcnt, x0, x1, yscale, lod=True):
         df = self.df.loc[x0:x1, :]
+        if self.is_sparse:
+            df = df.loc[df.iloc[:,self.col_data_offset].notna(), :]
+            if len(df) == 0:
+                # we need something that renders
+                df = self.df.loc[self.df.iloc[:,self.col_data_offset].notna(), :]
         origlen = len(df)
         return self._rows(df, colcnt, yscale=yscale, lod=lod), origlen
 
@@ -559,10 +577,16 @@ class FinLine(pg.GraphicsObject):
 
     def paint(self, p, *args):
         p.setPen(self.pen)
-        p.drawLine(QtCore.QPointF(*self.points[0]), QtCore.QPointF(*self.points[1]))
+        p.drawPath(self.shape())
+
+    def shape(self):
+        p = QtGui.QPainterPath()
+        p.moveTo(*self.points[0])
+        p.lineTo(*self.points[1])
+        return p
 
     def boundingRect(self):
-        return QtCore.QRectF(*self.points[0], *self.points[1])
+        return self.shape().boundingRect()
 
 
 class FinEllipse(pg.EllipseROI):
@@ -579,6 +603,7 @@ class FinViewBox(pg.ViewBox):
         self.v_zoom_scale = v_zoom_scale
         self.v_zoom_baseline = 0.5
         self.v_autozoom = True
+        self.max_zoom_points_f = 1
         self.y_max = 1000
         self.y_min = 0
         self.y_positive = True
@@ -757,7 +782,7 @@ class FinViewBox(pg.ViewBox):
             else: # sloppy one based on time stamps
                 tt0,tt1,_,_,_ = self.datasrc.hilo(tr.left(), tr.right())
                 vt0,vt1,_,_,_ = view.datasrc.hilo(vr.left(), vr.right())
-                period2 = self.datasrc.period * 0.5
+                period2 = self.datasrc.period * 0.5 * 1e9
                 if is_dirty or abs(vt0-tt0) >= period2 or abs(vt1-tt1) >= period2:
                     if is_dirty:
                         view.force_range_update -= 1
@@ -823,7 +848,8 @@ class FinViewBox(pg.ViewBox):
         # fetch hi-lo and set range
         _,_,hi,lo,cnt = datasrc.hilo(x0, x1)
         vr = self.viewRect()
-        if (x1-x0) < vr.width() and cnt < max_zoom_points:
+        minlen = int((max_zoom_points-0.5) * self.max_zoom_points_f + 0.51)
+        if (x1-x0) < vr.width() and cnt < minlen:
             return
         if not self.v_autozoom:
             hi = vr.bottom()
@@ -949,6 +975,7 @@ class CandlestickItem(FinPlotItem):
         self.draw_body = draw_body
         self.draw_shadow = draw_shadow
         self.candle_width = candle_width
+        self.shadow_width = 1
         self.colorfunc = colorfunc
         self.x_offset = 0
         super().__init__(ax, datasrc, lod=True)
@@ -967,7 +994,7 @@ class CandlestickItem(FinPlotItem):
             if self.x_offset:
                 idxs += self.x_offset
             if self.draw_shadow:
-                p.setPen(pg.mkPen(shadow))
+                p.setPen(pg.mkPen(shadow, width=self.shadow_width))
                 for x,(t,open,close,high,low) in zip(idxs, rows):
                     if high > low:
                         p.drawLine(QtCore.QPointF(x, low), QtCore.QPointF(x, high))
@@ -1184,7 +1211,7 @@ def create_plot_widget(master, rows=1, init_zoom_periods=1e10, yscale='linear'):
         ysc = yscale[n] if type(yscale) in (list,tuple) else yscale
         ysc = YScale(ysc, 1)
         v_zoom_scale = 0.97
-        viewbox = FinViewBox(master, init_steps=init_zoom_periods, yscale=ysc, v_zoom_scale=v_zoom_scale)
+        viewbox = FinViewBox(master, init_steps=init_zoom_periods, yscale=ysc, v_zoom_scale=v_zoom_scale, enableMenu=False)
         ax = prev_ax = _add_timestamp_plot(master=master, prev_ax=prev_ax, viewbox=viewbox, index=n, yscale=ysc)
         if axs:
             ax.setXLink(axs[0].vb)
@@ -1351,7 +1378,7 @@ def heatmap(datasrc, ax=None, **kwargs):
     if ax.vb.datasrc is not None and not ax.vb.datasrc.timebased(): # manual zoom update
         ax.setXLink(None)
         if ax.prev_ax:
-            ax.prev_ax.showAxis('bottom')
+            ax.prev_ax.set_visible(xaxis=True)
         df = ax.vb.datasrc.df
         prices = df.columns[ax.vb.datasrc.col_data_offset:item.col_data_end]
         delta_price = abs(prices[0] - prices[1])
@@ -1405,12 +1432,13 @@ def plot(x, y=None, color=None, width=1, ax=None, style=None, legend=None, zooms
         yfilter = y.notnull()
         ser = y.loc[yfilter]
         x = x.loc[yfilter].values if hasattr(x, 'loc') else x[yfilter]
-        item = ax.plot(x, ser.values, pen=None, symbol=symbol, symbolPen=None, symbolSize=5*width, symbolBrush=pg.mkBrush(used_color), name=legend)
+        item = ax.plot(x, ser.values, pen=None, symbol=symbol, symbolPen=None, symbolSize=7*width, symbolBrush=pg.mkBrush(used_color), name=legend)
         item.scatter._dopaint = item.scatter.paint
         item.scatter.paint = partial(_paint_scatter, item.scatter)
         # optimize (when having large number of points) by ignoring scatter click detection
         _dummy_mouse_click = lambda ev: 0
         item.scatter.mouseClickEvent = _dummy_mouse_click
+        item.setZValue(10)
     item.opts['handed_color'] = color
     item.ax = ax
     item.datasrc = datasrc
@@ -1572,6 +1600,7 @@ def show(qt_exec=True):
         if viewrestore:
             if _loadwindata(win):
                 continue
+        _set_max_zoom(vbs)
         for vb in vbs:
             datasrc = vb.datasrc_or_standalone
             if datasrc and (vb.linkedView(0) is None or vb.linkedView(0).datasrc is None):
@@ -1691,9 +1720,9 @@ def _add_timestamp_plot(master, prev_ax, viewbox, index, yscale):
     axes = {'bottom': EpochAxisItem(vb=viewbox, orientation='bottom'),
             'left':   YAxisItem(vb=viewbox, orientation='left')}
     if native_win:
-        ax = pg.PlotItem(viewBox=viewbox, axisItems=axes, name='plot-%i'%index)
+        ax = pg.PlotItem(viewBox=viewbox, axisItems=axes, name='plot-%i'%index, enableMenu=False)
     else:
-        axw = pg.PlotWidget(viewBox=viewbox, axisItems=axes, name='plot-%i'%index)
+        axw = pg.PlotWidget(viewBox=viewbox, axisItems=axes, name='plot-%i'%index, enableMenu=False)
         ax = axw.plotItem
         ax.ax_widget = axw
     ax.axes['left']['item'].setWidth(y_label_width) # this is to put all graphs on equal footing when texts vary from 0.4 to 2000000
@@ -1717,11 +1746,11 @@ def _add_timestamp_plot(master, prev_ax, viewbox, index, yscale):
     return ax
 
 
-def _overlay(ax, scale=0.25, y_axis=False):
+def _overlay(ax, scale=0.25, yaxis=False):
     '''The scale parameter defines how "high up" on the initial plot this overlay will show.
-       The y_axis parameter can be one of [False, 'linear', 'log'].'''
-    yscale = y_axis if y_axis else 'linear'
-    viewbox = FinViewBox(ax.vb.win, init_steps=ax.vb.init_steps, yscale=YScale(yscale, 1))
+       The yaxis parameter can be one of [False, 'linear', 'log'].'''
+    yscale = yaxis if yaxis else 'linear'
+    viewbox = FinViewBox(ax.vb.win, init_steps=ax.vb.init_steps, yscale=YScale(yscale, 1), enableMenu=False)
     viewbox.master_viewbox = ax.vb
     viewbox.setZValue(-5)
     viewbox.setBackgroundColor(ax.vb.state['background'])
@@ -1734,7 +1763,7 @@ def _overlay(ax, scale=0.25, y_axis=False):
     viewbox.setXLink(ax.vb)
     def updateView():
         viewbox.setGeometry(ax.vb.sceneBoundingRect())
-    axo = pg.PlotItem()
+    axo = pg.PlotItem(enableMenu=False)
     axo.significant_decimals = significant_decimals
     axo.significant_eps = significant_eps
     axo.vb = viewbox
@@ -1746,7 +1775,7 @@ def _overlay(ax, scale=0.25, y_axis=False):
     axo.hideAxis('bottom')
     axo.hideButtons()
     viewbox.addItem(axo)
-    if y_axis and isinstance(axo.vb.win, pg.GraphicsLayoutWidget):
+    if yaxis and isinstance(axo.vb.win, pg.GraphicsLayoutWidget):
         axi = YAxisItem(vb=axo.vb, orientation='right')
         axo.axes['right'] = {'item':axi}
         axi.linkToView(axo.vb)
@@ -1761,15 +1790,15 @@ def _overlay(ax, scale=0.25, y_axis=False):
     return axo
 
 
-def _ax_set_visible(ax, crosshair=None, xaxis=None, yaxis=None, yaxis_labels=None):
+def _ax_set_visible(ax, crosshair=None, xaxis=None, yaxis=None, xgrid=None, ygrid=None):
     if crosshair == False:
         ax.crosshair.hide()
     if xaxis is not None:
-        ax.showAxis('bottom', xaxis)
+        ax.getAxis('bottom').setStyle(showValues=xaxis)
     if yaxis is not None:
-        ax.showAxis('left', yaxis)
-    if yaxis_labels is not None:
-        ax.axes['left']['item'].hide_strings = not yaxis_labels
+        ax.getAxis('left').setStyle(showValues=yaxis)
+    if xgrid is not None or ygrid is not None:
+        ax.showGrid(x=xgrid, y=ygrid)
 
 
 def _ax_decouple(ax):
@@ -1814,9 +1843,8 @@ def _create_series(a):
     return a if isinstance(a, pd.Series) else pd.Series(a)
 
 
-def _create_datasrc(ax, *args, datacols=1):
-    def do_create(*args):
-        args = [a for a in args if a is not None]
+def _create_datasrc(ax, *args):
+    def do_create(args):
         if len(args) == 1 and type(args[0]) == PandasDataSource:
             return args[0]
         if len(args) == 1 and type(args[0]) in (list, tuple):
@@ -1827,9 +1855,10 @@ def _create_datasrc(ax, *args, datacols=1):
             return PandasDataSource(args[0])
         args = [_create_series(a) for a in args]
         return PandasDataSource(pd.concat(args, axis=1))
-    datasrc = do_create(*args)
+    iargs = [a for a in args if a is not None]
+    datasrc = do_create(iargs)
     # check if time column missing
-    if len(datasrc.df.columns) == datacols:
+    if len(datasrc.df.columns) == 1:
         # assume time data has already been added before
         for a in ax.vb.win.axs:
             if a.vb.datasrc and len(a.vb.datasrc.df.columns) >= 2:
@@ -1838,6 +1867,16 @@ def _create_datasrc(ax, *args, datacols=1):
                 datasrc.df.insert(0, col, a.vb.datasrc.df[col])
                 datasrc = PandasDataSource(datasrc.df)
                 break
+    elif len(iargs) >= 2 and len(datasrc.df.columns) == len(iargs)+1 and len(iargs) == len(args):
+        try:
+            if '.Int' in str(type(iargs[0].index)):
+                print('WARNING: performance penalty and crash may occur when using int64 instead of range indices.')
+                if (iargs[0].index == range(len(iargs[0]))).all():
+                    print(' - Fix by .reset_index(drop=True)')
+                    return _create_datasrc(ax, datasrc.df[datasrc.df.columns[1:]])
+        except:
+            print('WARNING: input data source may cause performance penalty and crash.')
+
     # FIX: stupid QT bug causes rectangles larger than 2G to flicker, so scale rendering down some
     if datasrc.df.iloc[:, 1:].max(numeric_only=True).max() > 1e8: # too close to 2G for comfort
         ax.vb.yscale.set_scale(int(1e8))
@@ -1879,6 +1918,22 @@ def _has_timecol(df):
     return len(df.columns) >= 2
 
 
+def _set_max_zoom(vbs):
+    '''Set the relative allowed zoom level between axes groups, where the lowest-resolution
+       plot in each group uses max_zoom_points, while the others get a scale >=1 of their
+       respective highest zoom level.'''
+    groups = defaultdict(set)
+    for vb in vbs:
+        master_vb = vb.linkedView(0)
+        if vb.datasrc and master_vb and master_vb.datasrc:
+            groups[master_vb].add(vb)
+            groups[master_vb].add(master_vb)
+    for group in groups.values():
+        minlen = min(len(vb.datasrc.df) for vb in group)
+        for vb in group:
+            vb.max_zoom_points_f = len(vb.datasrc.df) / minlen
+
+
 def _adjust_renko_datasrc(bins, step, datasrc):
     if not bins and not step:
         bins = 50
